#include <port_ipsec.h>

#include <linux/netdevice.h>
#include <linux/rtnetlink.h>
#include <net/protocol.h>

#include <net/icmp.h>

#include <net/raw.h>


#ifdef INET6
#include <net/rawv6.h>
#include <net/ipv6.h>

extern struct hdrtype_proc
{
       int     type;
       int     (*func) (struct sk_buff **, int offset);
} hdrproc_lst[];
#endif

/**
 *  skb_free_fraglist(): Frees the memory assigned to the fraglist of an skb.
 *
 */ 
void skb_free_fraglist(struct sk_buff *skb)
{
        struct sk_buff *list = skb_shinfo(skb)->frag_list;
        struct sk_buff *this = NULL;

        skb_shinfo(skb)->frag_list = NULL;

	for (this = list; this; this = list)
	{
		list = list->next;
		kfree_skb(this);	
	}
}


/**
 *  ipsec_skb_release_data(): Frees/Decreases the reference count of the data 
 *              buffer of a packet pointed to by an skb pointer.
 *
 */ 
void ipsec_skb_release_data(struct sk_buff *skb)
{
        int i;

        if (!skb->cloned || atomic_dec_and_test(&(skb_shinfo(skb)->dataref))) 
	{
                if (skb_shinfo(skb)->nr_frags) 
		{
                        for (i = 0; i < skb_shinfo(skb)->nr_frags; i++)
                                put_page(skb_shinfo(skb)->frags[i].page);
                }

                if (skb_shinfo(skb)->frag_list != NULL)
                        skb_free_fraglist(skb);

                kfree(skb->head);
        }
}


/**
 *  skb_copyback
 *      
 *  @skb: sk_buff 
 *  @offset: offset to start copy from.
 *  @len: length of data to be copied.
 *  @cp: buffer to copy from.
 *
 *  Copy data from a buffer back into the indicated sk_buff,
 *  starting "offset" bytes from the beginning, extending the mbuf
 *  chain if necessary.
 */
void skb_copyback(struct sk_buff *skb, int offset, int len, char *cp)
{
	
        int i, copy;
        int start = skb->len - skb->data_len;

        if (offset > (int)skb->len - len)
               return;

        /* Copy header. */
        if ((copy = start-offset) > 0) {
                if (copy > len)
                        copy = len;
                memcpy(skb->data + offset, cp, copy);
		
                if ((len -= copy) == 0)
                        return;

                offset += copy;
                cp += copy;
        }

        for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
                int end;

                end = start + skb_shinfo(skb)->frags[i].size;

                if ((copy = end-offset) > 0) {
                        u8 *vaddr;

                        if (copy > len)
                                copy = len;

                        vaddr = kmap_skb_frag(&skb_shinfo(skb)->frags[i]);
                        memcpy
			(vaddr + skb_shinfo(skb)->frags[i].page_offset +
			 offset - start, 
					cp, copy);
                        kunmap_skb_frag(vaddr);

                        if ((len -= copy) == 0)
                                return;

                        offset += copy;
                        cp += copy;
                }
                start = end;
        }

        if (skb_shinfo(skb)->frag_list) {
                struct sk_buff *list;

                for (list = skb_shinfo(skb)->frag_list; list; list=list->next) {
                        int end;

                        end = start + list->len;

                        if ((copy = end-offset) > 0) {
                                if (copy > len)
                                        copy = len;

                                skb_copyback(list, offset-start, copy, cp);

                                if ((len -= copy) == 0)
                                        return;

                                offset += copy;
                                cp += copy;
                        }
                        start = end;
                }
        }
	return;
}
